# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import json
import os
import sys
from typing import Optional

from datatailr import ACL, User


class Image:
    """
    Represents a container image for a job.
    The image is defined based on its' python dependencies and two 'build scripts' expressed as Dockerfile commands.
    All attributes can be initialized with either a string or a file name.
    """

    def __init__(
        self,
        acl: Optional[ACL] = None,
        python_version: str = "auto",
        python_requirements: str | list[str] = "",
        build_script_pre: str = "",
        build_script_post: str = "",
        branch_name: Optional[str] = None,
        commit_hash: Optional[str] = None,
        path_to_repo: Optional[str] = None,
        path_to_module: Optional[str] = None,
    ):
        self.python_version = python_version
        self.acl = acl
        self.python_requirements = python_requirements
        self.build_script_pre = build_script_pre
        self.build_script_post = build_script_post
        self.branch_name = branch_name
        self.commit_hash = commit_hash
        self.path_to_repo = path_to_repo
        self.path_to_module = path_to_module

    def __repr__(self):
        return f"Image(acl={self.acl},)"

    @property
    def python_version(self):
        return self._python_version

    @python_version.setter
    def python_version(self, value: str):
        if not isinstance(value, str):
            raise TypeError("python_version must be a string.")
        if value.lower() == "auto":
            value = f"{sys.version_info.major}.{sys.version_info.minor}"
        if value not in ["3.10", "3.11", "3.12", "3.13", "3.14"]:
            raise ValueError(f"Invalid python_version: {value}")
        self._python_version = value

    @property
    def python_requirements(self):
        return self._python_requirements

    @python_requirements.setter
    def python_requirements(self, value: str | list[str]):
        if isinstance(value, str) and os.path.isfile(value):
            with open(value, "r") as f:
                value = f.read()
        elif isinstance(value, list):
            value = "\n".join(value)
        if not isinstance(value, str):
            raise TypeError(
                "python_requirements must be a string or a file path to a requirements file."
            )
        self._python_requirements = value

    @property
    def build_script_pre(self):
        return self._build_script_pre

    @build_script_pre.setter
    def build_script_pre(self, value: str):
        if not isinstance(value, str):
            raise TypeError(
                "build_script_pre must be a string or a file path to a script file."
            )
        if os.path.isfile(value):
            with open(value, "r") as f:
                value = f.read()
        self._build_script_pre = value

    @property
    def build_script_post(self):
        return self._build_script_post

    @build_script_post.setter
    def build_script_post(self, value: str):
        if not isinstance(value, str):
            raise TypeError(
                "build_script_post must be a string or a file path to a script file."
            )
        if os.path.isfile(value):
            with open(value, "r") as f:
                value = f.read()
        self._build_script_post = value

    @property
    def acl(self):
        return self._acl

    @acl.setter
    def acl(self, value: Optional[ACL]):
        if value is None:
            signed_user = User.signed_user()
            if signed_user is None:
                raise ValueError(
                    "ACL cannot be None. Please provide a valid ACL or ensure a user is signed in."
                )
        elif not isinstance(value, ACL):
            raise TypeError("acl must be an instance of ACL.")
        self._acl = value or ACL(signed_user)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            if (
                key in ["branch_name", "commit_hash", "path_to_repo", "path_to_module"]
                and value is not None
                and not isinstance(value, str)
            ):
                raise TypeError(f"{key} must be a string or None.")
            if not hasattr(self, key):
                raise AttributeError(
                    f"'{self.__class__.__name__}' object has no attribute '{key}'"
                )
            setattr(self, key, value)

    def to_dict(self):
        """
        Convert the Image instance to a dictionary representation.
        """
        return {
            "acl": self.acl.to_dict(),
            "python_version": self.python_version,
            "python_requirements": self.python_requirements,
            "build_script_pre": self.build_script_pre,
            "build_script_post": self.build_script_post,
            "branch_name": self.branch_name,
            "commit_hash": self.commit_hash,
            "path_to_repo": self.path_to_repo,
            "path_to_module": self.path_to_module,
        }

    def to_json(self):
        """
        Convert the Image instance to a JSON string representation.
        """
        return json.dumps(self.to_dict(), indent=4)



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from typing import Optional
from datatailr import User
from datatailr.wrapper import dt__Tag


def get_log_level() -> int:
    log_level = os.getenv("DATATAILR_LOG_LEVEL", "INFO").upper()
    if log_level in ["TRACE", "DEBUG"]:
        return logging.DEBUG
    elif log_level == "INFO":
        return logging.INFO
    elif log_level == "WARNING":
        return logging.WARNING
    elif log_level == "ERROR":
        return logging.ERROR
    elif log_level == "CRITICAL":
        return logging.CRITICAL
    else:
        return logging.INFO


def ansi_symbols_supported() -> bool:
    """Check if the terminal supports ANSI symbols."""
    if sys.platform.startswith("win"):
        return (
            os.getenv("ANSICON") is not None
            or os.getenv("WT_SESSION") is not None
            or "TERM" in os.environ
            and os.environ["TERM"] == "xterm-256color"
        )
    else:
        return sys.stdout.isatty()


ANSI_AVAILABLE = ansi_symbols_supported()


def color_text(text: str, color_name: str) -> str:
    """Wrap text with ANSI color codes if supported."""
    if not ANSI_AVAILABLE:
        return text

    colors = {
        "red": "\033[31m",
        "green": "\033[32m",
        "yellow": "\033[33m",
        "blue": "\033[34m",
        "magenta": "\033[35m",
        "cyan": "\033[36m",
        "bold": "\033[1m",
        "reset": "\033[0m",
    }
    color_code = colors.get(color_name.lower(), "")
    reset_code = colors["reset"] if color_code else ""
    return f"{color_code}{text}{reset_code}"


def RED(text: str) -> str:
    return color_text(text, "red")


def GREEN(text: str) -> str:
    return color_text(text, "green")


def YELLOW(text: str) -> str:
    return color_text(text, "yellow")


def BLUE(text: str) -> str:
    return color_text(text, "blue")


def MAGENTA(text: str) -> str:
    return color_text(text, "magenta")


def CYAN(text: str) -> str:
    return color_text(text, "cyan")


def BOLD(text: str) -> str:
    return color_text(text, "bold")


class MaxLevelFilter(logging.Filter):
    """Allow only log records at or below a given level."""

    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno <= self.level


class MinLevelFilter(logging.Filter):
    """Allow only log records at or above a given level."""

    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno >= self.level


class ColoredFormatter(logging.Formatter):
    COLORS = {
        logging.DEBUG: "\033[34m",  # Blue
        logging.INFO: "\033[32m",  # Green
        logging.WARNING: "\033[33m",  # Yellow
        logging.ERROR: "\033[31m",  # Red
        logging.CRITICAL: "\033[41m",  # Red background
    }
    RESET = "\033[0m"
    BOLD = "\033[1m"

    def format(self, record):
        color = self.COLORS.get(record.levelno, self.RESET)
        timestamp = f"{self.BOLD}{self.formatTime(record)}{self.RESET}"
        level = f"{color}{record.levelname}{self.RESET}"
        message = f"{color}{record.getMessage()}{self.RESET}"
        LOG_FORMAT = f"{timestamp} - {level} - {node_name}:{node_ip} - {user} - {job_name} - {record.name} - [Line {record.lineno}]: {message}"
        return LOG_FORMAT


tag = dt__Tag()
node_name = tag.get("node_name") or "local"
node_ip = tag.get("node_ip")
job_name = os.getenv("DATATAILR_JOB_NAME", "unknown_job")

try:
    user = User.signed_user().name
except Exception:
    import getpass

    user = getpass.getuser()


class DatatailrLogger:
    def __init__(
        self,
        name: str,
        log_file: Optional[str] = None,
        log_level: int = get_log_level(),
    ):
        """
        Initialize the DatatailrLogger.

        :param name: Name of the logger.
        :param log_file: Optional file to log messages to.
        :param log_level: Logging level (default: logging.INFO).
        """
        self.logger = logging.getLogger(name)
        self.logger.setLevel(log_level)

        formatter = ColoredFormatter()

        # stdout handler (DEBUG/INFO only)
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.addFilter(MaxLevelFilter(logging.INFO))
        stdout_handler.setFormatter(formatter)
        self.logger.addHandler(stdout_handler)

        # stderr handler (WARNING and above)
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setLevel(logging.WARNING)
        stderr_handler.addFilter(MinLevelFilter(logging.WARNING))
        stderr_handler.setFormatter(formatter)
        self.logger.addHandler(stderr_handler)

        # Optional file handler
        if log_file:
            file_handler = RotatingFileHandler(
                log_file, maxBytes=10 * 1024 * 1024, backupCount=5
            )
            file_handler.setLevel(log_level)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

        self.enable_opentelemetry()

    def get_logger(self):
        """
        Get the configured logger instance.

        :return: Configured logger.
        """
        return self.logger

    def enable_opentelemetry(self):
        """
        Enable OpenTelemetry integration if available.
        """
        try:
            from opentelemetry.instrumentation.logging import LoggingInstrumentor  # type: ignore

            LoggingInstrumentor().instrument(set_logging_format=True)
        except ImportError:
            pass  # OpenTelemetry is not installed, skip instrumentation



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import functools
import inspect
import os
from typing import Callable

from datatailr.logging import DatatailrLogger
from datatailr.scheduler.arguments_cache import ArgumentsCache, CacheNotFoundError
from datatailr.scheduler.base import EntryPoint, JobType, Resources
from datatailr.scheduler.batch import (
    BatchJob,
    get_current_manager,
)
from datatailr.scheduler.constants import DEFAULT_TASK_CPU, DEFAULT_TASK_MEMORY
from datatailr.scheduler.utils import get_available_env_args


__FUNCTIONS_CREATED_IN_DAG__: dict[Callable, str] = {}
logger = DatatailrLogger(__name__).get_logger()

__ARGUMENTS_CACHE__ = None


def get_arguments_cache() -> ArgumentsCache:
    global __ARGUMENTS_CACHE__
    if __ARGUMENTS_CACHE__ is None:
        __ARGUMENTS_CACHE__ = ArgumentsCache()
    return __ARGUMENTS_CACHE__


def batch_run_id() -> str:
    return os.getenv("DATATAILR_BATCH_RUN_ID", "unknown")


def dag_id(job: BatchJob) -> str:
    return os.getenv(
        "DATATAILR_BATCH_ID", __FUNCTIONS_CREATED_IN_DAG__.get(job, "unknown")
    )


def batch_decorator(memory: str = DEFAULT_TASK_MEMORY, cpu: float = DEFAULT_TASK_CPU):
    """
    Decorator to mark a function as a batch job.
    This decorator can be used to wrap functions that should be executed as part of batch jobs.
    """

    def decorator(func) -> BatchJob:
        spec = inspect.getfullargspec(func)
        signature = inspect.signature(func)
        varargs = spec.varargs
        varkw = spec.varkw
        parameters = signature.parameters

        @functools.wraps(func)
        def batch_main(*args, **kwargs):
            dag = get_current_manager()
            if dag is None:
                logger.info(f'Function "{func.__name__}" is being executed.')
                # There are two possible scenarios:
                # 1. The function is called directly, not as part of a batch job. In this case, the args and kwargs should be used.
                # 2. The function is called as part of a batch job - it was constructed as part of a DAG and is now being executed.
                env_args = get_available_env_args()
                all_function_args = [
                    v.name
                    for v in parameters.values()
                    if v.kind
                    not in (
                        inspect.Parameter.VAR_POSITIONAL,
                        inspect.Parameter.VAR_KEYWORD,
                    )
                ]
                final_args = list(args)

                for name, value in env_args.items():
                    if name in all_function_args:
                        if len(final_args) < len(all_function_args):
                            final_args.extend(
                                [None] * (len(all_function_args) - len(final_args))
                            )
                        final_args[all_function_args.index(name)] = value
                try:
                    final_kwargs = get_arguments_cache().get_arguments(
                        dag_id(func),
                        os.getenv("DATATAILR_JOB_NAME", func.__name__),
                        os.getenv("DATATAILR_BATCH_RUN_ID"),
                    )
                except CacheNotFoundError:
                    final_kwargs = kwargs

                if varargs is not None and varkw is None:
                    for key in list(final_kwargs.keys()):
                        if key not in parameters:
                            final_args.append(final_kwargs.pop(key))

                # Some of the loaded arguments are actually args and not kwargs.
                if len(final_args) == len(parameters.keys()):
                    for i, arg_name in enumerate(parameters.keys()):
                        final_args[i] = final_kwargs.pop(arg_name, final_args[i])
                result = func(*final_args, **final_kwargs)
                get_arguments_cache().add_result(
                    batch_run_id(),
                    os.getenv("DATATAILR_JOB_NAME", func.__name__),
                    result,
                )
                return result
            else:
                if varargs is not None:
                    all_args = {job.name: job for job in args}
                else:
                    all_args = dict(zip(spec.args, args)) | kwargs
                dag.set_autorun(True)

                job = BatchJob(
                    name=func.__name__,
                    entrypoint=EntryPoint(
                        JobType.BATCH,
                        func=func,
                    ),
                    resources=Resources(memory=memory, cpu=cpu),
                    dependencies=[
                        value.name
                        for _, value in all_args.items()
                        if isinstance(value, BatchJob)
                    ],
                    dag=dag,
                )
                job.args = all_args
                __FUNCTIONS_CREATED_IN_DAG__[job.entrypoint.func] = dag.id
                return job

        module = inspect.getmodule(func)
        if hasattr(module, "__batch_main__"):
            if func.__name__ in getattr(module, "__batch_main__"):
                raise ValueError(f"Duplicate batch main function {func.__name__}")
            module.__batch_main__[func.__name__] = batch_main  # type: ignore
        else:
            setattr(module, "__batch_main__", {func.__name__: batch_main})

        # The return type is a BatchJob, but we use type: ignore to avoid type checking issues
        return batch_main  # type: ignore

    return decorator



from typing import Callable, Optional, Dict, Union
import functools
from datatailr.scheduler import Batch, Schedule, Resources
from datatailr import Environment, Image, ACL, User


def workflow(
    name: str,
    schedule: Optional[Schedule] = None,
    image: Optional[Image] = None,
    run_as: Optional[Union[str, User]] = None,
    resources: Resources = Resources(memory="100m", cpu=1),
    acl: Optional[ACL] = None,
    python_version: str = "3.12",
    python_requirements: str = "",
    build_script_pre: str = "",
    build_script_post: str = "",
    env_vars: Dict[str, str | int | float | bool] = {},
):
    _name = name
    _schedule = schedule
    _image = image
    _run_as = run_as
    _resources = resources
    _acl = acl
    _python_version = python_version
    _python_requirements = python_requirements
    _build_script_pre = build_script_pre
    _build_script_post = build_script_post
    _env_vars = env_vars

    def decorator(func) -> Callable:
        @functools.wraps(func)
        def wrapper(
            *args,
            local_run: bool = False,
            schedule: Optional[Schedule] = None,
            image: Optional[Image] = None,
            run_as: Optional[Union[str, User]] = None,
            resources: Optional[Resources] = None,
            acl: Optional[ACL] = None,
            python_version: Optional[str] = None,
            python_requirements: Optional[str] = None,
            build_script_pre: Optional[str] = None,
            build_script_post: Optional[str] = None,
            env_vars: Optional[Dict[str, str | int | float | bool]] = None,
            **kwargs,
        ):
            __schedule = schedule or _schedule
            __image = image or _image
            __run_as = run_as or _run_as
            __resources = resources or _resources
            __acl = acl or _acl
            __python_version = python_version or _python_version
            __python_requirements = python_requirements or _python_requirements
            __build_script_pre = build_script_pre or _build_script_pre
            __build_script_post = build_script_post or _build_script_post
            __env_vars = env_vars or _env_vars

            if local_run and (__schedule is not None):
                raise ValueError("Cannot set schedule for local run.")

            dag = Batch(
                name=_name,
                environment=Environment.DEV,
                schedule=__schedule,
                image=__image,
                run_as=__run_as,
                resources=__resources,
                acl=__acl,
                local_run=local_run,
                python_version=__python_version,
                python_requirements=__python_requirements,
                build_script_pre=__build_script_pre,
                build_script_post=__build_script_post,
                env_vars=__env_vars,
            )
            dag.set_autorun(False)
            with dag:
                func(*args, **kwargs)

        return wrapper

    return decorator



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

"""Datatailr demo project template package.

Use the console script `datatailr init-demo` to copy these template files
into a target directory where you can experiment.
"""



#!/usr/bin/env python3

# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

import os
import subprocess

from datatailr.logging import DatatailrLogger

logger = DatatailrLogger(os.path.abspath(__file__)).get_logger()


def run():
    logger.info("Starting Datatailr excel add-in...")

    entrypoint = os.environ.get("DATATAILR_ENTRYPOINT")
    if entrypoint is None:
        raise ValueError("Environment variable 'DATATAILR_ENTRYPOINT' is not set.")

    local = os.environ.get("DATATAILR_LOCAL", False) in ("1", "true", "True", "TRUE")
    if local:
        hostname = "localhost.excel.datatailr.com"
        local_flag = "-l"
    else:
        hostname = os.environ.get("EXCEL_HOST")
        if hostname is None:
            raise ValueError("Environment variable 'EXCEL_HOST' is not set.")
        local_flag = ""

    module_name = entrypoint.split(":", 1)[0]

    entrypoint = f'/opt/datatailr/bin/dt-excel.sh -n -H "{hostname}" {local_flag} -p 8080 -w 8000 {module_name}'
    logger.info(f"Running entrypoint: {entrypoint}")
    subprocess.run(entrypoint, shell=True)



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from typing import Any, ClassVar


class dt__Tag:
    """
    Tag management for local runs in the absence of the DataTailr platform.
    All instances share the same tag store.
    """

    # shared across all instances
    tags: ClassVar[dict[str, Any]] = {
        "blob_storage_prefix": "local-no-dt-",
    }

    def ls(self) -> dict[str, Any]:
        return self.__class__.tags

    def get(self, name: str) -> Any:
        return self.__class__.tags.get(name)

    def set(self, name: str, value: Any) -> None:
        self.__class__.tags[name] = value

    def rm(self, name: str) -> None:
        self.__class__.tags.pop(name, None)



# type: ignore

# noqa: F401
try:
    from datatailr.excel.addin import Addin
    from datatailr.excel.addin import Queue
except ImportError:

    class DummyAddin:
        def __init__(self, name: str, description: str) -> None:
            self.name = name
            self.description = description

        def expose(self, **kwargs):
            def decorator(func):
                return func

            return decorator

        def run(self):
            pass

    class DummyQueue:
        def __init__(self) -> None:
            pass

        def push(self, value) -> None:
            pass

    Addin = DummyAddin
    Queue = DummyQueue


__all__ = ["Addin", "Queue"]



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

DEFAULT_TASK_MEMORY = "128m"
DEFAULT_TASK_CPU = 0.25

BATCH_JOB_ARGUMENTS = (
    "rundate",
    "scheduled_time",
    "started_at",
    "batch_name",
    "job_name",
)



#!/usr/bin/env python3

# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************


# The purpose of this script is to be the entrypoint for all jobs running on datatailr.
# The main functions of the script are:
#     1. Create a linux user and group for the job.
#     2. Set the environment variables for the job.
#     3. Run the job in a separate process, as the newly created user and pass all relevant environment variables.
# There are muliple environment variables which are required for the job to run.
# Some of them are necessary for the setup stage, which is executed directly in this script as the linux root user.
# Others are passed to the job script, which is executed in a separate process with only the users' privileges and not as a root user.
#
# Setup environment variables:
#     DATATAILR_USER - the user under which the job will run.
#     DATATAILR_GROUP - the group under which the job will run.
#     DATATAILR_UID - the user ID of the user as it is defined in the system.
#     DATATAILR_GID - the group ID of the group as it is defined in the system.
#     DATATAILR_JOB_TYPE - the type of job to run. (batch\service\app\excel\IDE)
# Job environment variables (not all are always relevant, depending on the job type):
#     DATATAILR_BATCH_RUN_ID - the unique identifier for the batch run.
#     DATATAILR_BATCH_ID - the unique identifier for the batch.
#     DATATAILR_JOB_ID - the unique identifier for the job.


import concurrent.futures
import subprocess
import os
import sys
import shlex
import shutil
import sysconfig
from typing import Optional, Tuple
from datatailr.logging import DatatailrLogger
from datatailr.utils import is_dt_installed
from datatailr import User
from pathlib import Path
import importlib.resources as ir

logger = DatatailrLogger(os.path.abspath(__file__)).get_logger()

if not is_dt_installed():
    logger.error("Datatailr is not installed.")
    sys.exit(1)


def get_env_var(name: str, default: str | None = None) -> str:
    """
    Get an environment variable.
    If the variable is not set, raise an error.
    """
    if name not in os.environ:
        if default is not None:
            return default
        logger.error(f"Environment variable '{name}' is not set.")
        raise ValueError(f"Environment variable '{name}' is not set.")
    return os.environ[name]


def get_python_scripts_location() -> str:
    """
    Get the location of the python scripts.
    """
    scripts_path_file_name = "/opt/datatailr/etc/python-scripts-location"
    if os.path.exists(scripts_path_file_name):
        with open(scripts_path_file_name, "r") as f:
            scripts_path = f.read().strip()
            return scripts_path
    return ""


def create_user_and_group() -> Tuple[str, str]:
    """
    Create a user and group for the job.
    The user and group names are taken from the environment variables DATATAILR_USER and DATATAILR_GROUP.
    The group and user are created with the same uid and gid as passed in the environment variables DATATAILR_UID and DATATAILR_GID.
    If the user or group already exists, do nothing.
    """
    user = get_env_var("DATATAILR_USER")
    group = get_env_var("DATATAILR_GROUP")
    uid = get_env_var("DATATAILR_UID")
    gid = get_env_var("DATATAILR_GID")

    # Create group if it does not exist
    # -o: allow to create a group with a non-unique GID
    os.system(f"getent group {group} || groupadd {group} -g {gid} -o")

    # Create user if it does not exist
    # -s /bin/bash: set the shell to bash
    # -o: allow to create a user with a non-unique UID
    os.system(
        f"getent passwd {user} || useradd -g {group} -s /bin/bash -u {uid} -o {user}"
    )

    for directory in [
        ".cache",
        ".config",
        ".config/code-server",
        ".local",
        "tmp",
        "tmp/.dt",
        ".vscode",
    ]:
        dir_name = f"/home/{user}/{directory}"
        if os.path.exists(dir_name):
            continue
        os.makedirs(dir_name, exist_ok=True)
        os.chown(dir_name, int(uid), int(gid))
        os.chmod(dir_name, 0o773)  # rwxrwxr-x

    copy_dotfiles(user, int(uid), int(gid))
    create_git_config(user)
    copy_welcome_readme(user)
    return user, group


def copy_dotfiles(user: str, uid: int, gid: int):
    home_dir = f"/home/{user}"
    skel_dir = "/etc/skel"
    # Copy dotfiles from /etc/skel to the new user's home directory if they do not already exist
    for root, dirs, files in os.walk(skel_dir):
        rel_path = os.path.relpath(root, skel_dir)
        target_root = os.path.join(home_dir, rel_path) if rel_path != "." else home_dir

        for d in dirs:
            target_dir = os.path.join(target_root, d)
            if not os.path.exists(target_dir):
                os.makedirs(target_dir, exist_ok=True)
                os.chown(target_dir, uid, gid)  # root:root
                os.chmod(target_dir, 0o773)  # rwxrwxr-x

        for f in files:
            src_file = os.path.join(root, f)
            target_file = os.path.join(target_root, f)
            if not os.path.exists(target_file):
                shutil.copy(src_file, target_file)
                os.chown(target_file, uid, gid)
                os.chmod(target_file, 0o773)  # rwxrwxr-x


def create_git_config(user: str):
    dt_user = User(user)
    git_user_email = subprocess.run(
        ["sudo", "-u", user, "git", "config", "--global", "user.email"],
        capture_output=True,
        text=True,
    ).stdout.strip()

    if git_user_email not in ["None", ""]:
        return

    for key, value in [
        ("user.name", f"{dt_user.first_name} {dt_user.last_name}"),
        ("user.email", f"{dt_user.email}"),
        ("init.defaultBranch", "main"),
    ]:
        subprocess.run(["sudo", "-u", user, "git", "config", "--global", key, value])
    logger.info(f"Git global config created for user {user}.")


def link_script_files():
    scripts_location = get_python_scripts_location()
    if not scripts_location or scripts_location == "":
        return
    user_bin_dir = "/usr/bin"
    script_files = ["datatailr"]
    for script_file in script_files:
        src_path = os.path.join(scripts_location, script_file)
        if not os.path.exists(src_path):
            logger.warning(f"Script {script_file} not found in {scripts_location}.")
            continue
        dest_path = os.path.join(user_bin_dir, script_file)
        if not os.path.exists(dest_path):
            os.symlink(src_path, dest_path)
            logger.info(f"Linked script {script_file} to {dest_path}")


def copy_welcome_readme(user: str):
    demo_package_root = Path(str(ir.files("datatailr_demo")))
    readme_src = demo_package_root / "README.md"
    if not readme_src.exists():
        logger.warning(
            f"Demo README.md not found in datatailr_demo package at: {readme_src}"
        )
        return
    readme_dest = f"/home/{user}/README.md"
    if not os.path.exists(readme_dest):
        shutil.copy(readme_src, readme_dest)
        os.chown(readme_dest, 0, 0)
        os.chmod(readme_dest, 0o644)  # rw-r--r--


def prepare_command_argv(command: str | list, user: str, env_vars: dict) -> list[str]:
    if isinstance(command, str):
        command = shlex.split(command)

    python_libdir = sysconfig.get_config_var("LIBDIR")
    ld_library_path = get_env_var("LD_LIBRARY_PATH", None)

    # Base environment variables setup
    base_env = {
        "PATH": get_env_var("PATH", ""),
        "PYTHONPATH": get_env_var("PYTHONPATH", ""),
        "LD_LIBRARY_PATH": ":".join(filter(None, [python_libdir, ld_library_path])),
    }

    merged_env = base_env | env_vars
    env_kv = [f"{k}={v}" for k, v in merged_env.items()]
    return ["sudo", "-u", user, "env", *env_kv, *command]


def run_single_command_non_blocking(
    command: str | list,
    user: str,
    env_vars: dict,
    log_stream_name: Optional[str | None] = None,
) -> int:
    """
    Runs a single command non-blocking and returns the exit code after it finishes.
    This is designed to be run within an Executor.
    """
    argv = prepare_command_argv(command, user, env_vars)
    cmd_label = " ".join(argv[4:])  # For logging purposes

    try:
        if log_stream_name:
            stdout_file_path = f"/opt/datatailr/var/log/{log_stream_name}.log"
            stderr_file_path = f"/opt/datatailr/var/log/{log_stream_name}_error.log"
            with (
                open(stdout_file_path, "ab", buffering=0) as stdout_file,
                open(stderr_file_path, "ab", buffering=0) as stderr_file,
            ):
                proc = subprocess.Popen(argv, stdout=stdout_file, stderr=stderr_file)
        else:
            proc = subprocess.Popen(argv)
        returncode = proc.wait()

        if returncode != 0:
            logger.error(f"Command '{cmd_label}' failed with exit code {returncode}")
        return returncode
    except Exception as e:
        logger.error(f"Execution error for '{cmd_label}': {e}")
        return 1


def run_commands_in_parallel(
    commands: list[str | list],
    user: str,
    env_vars: dict,
    log_stream_names: Optional[list[str | None]] = None,
) -> int:
    """
    Executes two commands concurrently using a ThreadPoolExecutor.
    Returns a tuple of (return_code_cmd1, return_code_cmd2).
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=len(commands)) as executor:
        futures = []
        for command, log_stream_name in zip(
            commands, log_stream_names or [None] * len(commands)
        ):
            futures.append(
                executor.submit(
                    run_single_command_non_blocking,
                    command,
                    user,
                    env_vars,
                    log_stream_name,
                )
            )
        results = [
            future.result() for future in concurrent.futures.as_completed(futures)
        ]
        return 0 if all(code == 0 for code in results) else 1


def main():
    user, _ = create_user_and_group()
    job_type = get_env_var("DATATAILR_JOB_TYPE")

    env = {
        "DATATAILR_JOB_TYPE": job_type,
        "DATATAILR_JOB_NAME": get_env_var("DATATAILR_JOB_NAME"),
        "DATATAILR_JOB_ID": get_env_var("DATATAILR_JOB_ID"),
    }

    if job_type == "batch":
        run_id = get_env_var("DATATAILR_BATCH_RUN_ID")
        batch_id = get_env_var("DATATAILR_BATCH_ID")
        entrypoint = get_env_var("DATATAILR_BATCH_ENTRYPOINT")
        batch_args = {
            key: value
            for key, value in os.environ.items()
            if key.startswith("DATATAILR_BATCH_ARG_")
        }
        env = (
            {
                "DATATAILR_BATCH_RUN_ID": run_id,
                "DATATAILR_BATCH_ID": batch_id,
                "DATATAILR_BATCH_ENTRYPOINT": entrypoint,
            }
            | env
            | batch_args
        )
        return run_single_command_non_blocking("datatailr_run_batch", user, env)
    elif job_type == "service":
        port = get_env_var("DATATAILR_SERVICE_PORT", 8080)
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        env = {
            "DATATAILR_ENTRYPOINT": entrypoint,
            "DATATAILR_SERVICE_PORT": port,
        } | env
        return run_single_command_non_blocking("datatailr_run_service", user, env)
    elif job_type == "app":
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        env = {
            "DATATAILR_ENTRYPOINT": entrypoint,
        } | env
        return run_single_command_non_blocking("datatailr_run_app", user, env)
    elif job_type == "excel":
        host = get_env_var("EXCEL_HOST", "")
        entrypoint = get_env_var("DATATAILR_ENTRYPOINT")
        local = get_env_var("DATATAILR_LOCAL", False)
        dt_environment = get_env_var("DATATAILR_JOB_ENVIRONMENT")
        env = {
            "DATATAILR_ENTRYPOINT": entrypoint,
            "EXCEL_HOST": host,
            "DATATAILR_LOCAL": local,
            "DATATAILR_JOB_ENVIRONMENT": dt_environment,
        } | env
        return run_single_command_non_blocking("datatailr_run_excel", user, env)

    elif job_type == "workspace":
        os.makedirs("/opt/datatailr/var/log", exist_ok=True)
        ide_command = [
            "code-server",
            "--disable-telemetry",
            "--disable-update-check",
            "--auth=none",
            "--bind-addr=127.0.0.1:9090",
            f'--app-name="Datatailr IDE {get_env_var("DATATAILR_USER")}"',
            f"--user-data-dir=/home/{user}/.vscode",
            f"/home/{user}",
        ]
        job_name = get_env_var("DATATAILR_JOB_NAME")
        jupyter_command = [
            "jupyter-lab",
            "--ip='127.0.0.1'",
            "--port=7070",
            "--no-browser",
            "--NotebookApp.token=''",
            "--NotebookApp.password=''",
            f"--ServerApp.base_url=/workspace/dev/{job_name}/jupyter/",
            f"--ServerApp.file_url_prefix=/workspace/dev/{job_name}/jupyter/static/",
            f"--ServerApp.root_dir=/home/{user}",
        ]
        return run_commands_in_parallel(
            [ide_command, jupyter_command], user, env, ["code-server", "jupyter"]
        )

    else:
        raise ValueError(f"Unknown job type: {job_type}")


if __name__ == "__main__":
    try:
        logger.debug("Starting job execution...")
        rc = main()
        logger.debug(f"Job executed successfully, with code {rc}")
        raise SystemExit(rc)
    except Exception as e:
        logger.error(f"Error during job execution: {e}")
        raise SystemExit(1)



"""
Copyright (c) 2025 - Datatailr Inc.
All Rights Reserved.

This file is part of Datatailr and subject to the terms and conditions
defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
of this file, in parts or full, via any medium is strictly prohibited.
"""


class AddinBase:
    def __init__(self, name, *args, **kwargs):
        self.name = name

    def decorator_impl(
        self,
        signature,
        wrapper,
        func_name,
        description,
        help,
        volatile,
        streaming,
    ):
        pass


class Queue:
    def __init__(self, name, _id):
        self.name = name
        self.id = _id

    def push(self, value):
        print(f"Queue {self.name} ({self.id}): {value}")

    def error(self, message):
        print(f"Queue {self.name} ({self.id}) Error: {message}")



#!/usr/bin/env python3

# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************

from __future__ import annotations

import sys
import shutil
import os
from pathlib import Path
import subprocess
import argparse
import importlib.resources as ir
from datatailr.logging import CYAN, RED, YELLOW, GREEN
from datatailr import User

DEFAULT_TARGET = "datatailr_demo_project"


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="datatailr",
        description="Datatailr command line utilities. Use subcommands to access features.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    sub = parser.add_subparsers(dest="command", metavar="<command>")

    # init-demo subcommand
    p_init = sub.add_parser(
        "init-demo",
        help="Copy the bundled demo project into a target directory.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    p_init.add_argument(
        "-d",
        "--destination",
        metavar="PATH",
        default=str(Path.cwd() / DEFAULT_TARGET),
        help="Destination directory to create (will be created if missing).",
    )
    p_init.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Overwrite existing files (by default existing files are skipped).",
    )

    # version subcommand
    sub.add_parser("version", help="Show python and datatailr versions.")

    return parser


def copy_tree(src: Path, dst: Path, username: str, overwrite: bool = False) -> None:
    file_types = (
        ".py",
        ".yaml",
        ".yml",
        ".md",
        ".txt",
        ".csv",
        ".json",
        ".png",
        ".jpg",
        ".jpeg",
        ".ico",
    )
    for root, dirs, files in os.walk(src):
        rel = Path(root).relative_to(src)
        target_root = dst / rel
        target_root.mkdir(parents=True, exist_ok=True)
        for d in dirs:
            (target_root / d).mkdir(exist_ok=True)
        for f in files:
            if not f.lower().endswith(file_types):
                continue  # skip non-relevant files
            src_file = Path(root) / f
            dst_file = target_root / f
            if dst_file.exists() and not overwrite:
                continue  # skip existing unless force enabled
            # replace <>USERNAME<> placeholders in the files:

            with open(src_file, "r") as sf, open(dst_file, "w") as df:
                content = sf.read()
                content = content.replace("<>USERNAME<>", username)
                df.write(content)


def git_init_repo(path: Path) -> None:
    """Initialize a git repository at the given path if not already a git repo."""
    if (path / ".git").exists():
        return  # already a git repo

    if not shutil.which("git"):
        print(
            YELLOW(
                "Git is not installed or not found in PATH; skipping git initialization."
            )
        )
        return
    subprocess.run(["git", "init"], cwd=str(path), check=True)
    subprocess.run(["git", "add", "."], cwd=str(path), check=True)
    subprocess.run(["git", "commit", "-m", "Initial commit"], cwd=str(path), check=True)


def install_code_extensions() -> None:
    """Install recommended code-server extensions for Datatailr."""
    if not shutil.which("code-server"):
        print(
            YELLOW(
                "Code server is not installed or not found in PATH; skipping extension installation."
            )
        )
        return
    extensions = [
        "ms-python.python",
        "ms-python.debugpy",
    ]
    print(CYAN("Installing recommended code-server extensions..."))
    for ext in extensions:
        subprocess.run(["code-server", "--install-extension", ext], check=True)
    print(GREEN("Extensions installed."))


def open_folder(path: Path) -> None:
    """Launch code server in the target path."""
    if not shutil.which("code-server"):
        print(
            YELLOW(
                "Code server is not installed or not found in PATH; skipping code server launch."
            )
        )
        return
    subprocess.run(["code-server", str(path)], check=True)


def main(argv: list[str] | None = None) -> int:
    argv = argv or sys.argv[1:]
    parser = build_parser()
    args = parser.parse_args(argv)

    if args.command == "init-demo":
        try:
            template_root = ir.files("datatailr_demo")
        except Exception as e:  # pragma: no cover
            print(RED(f"Could not access demo template package: {e}"))
            return 1
        src_path = Path(str(template_root))

        target = Path(args.destination)
        if target.exists() and any(target.iterdir()) and not args.force:
            print(
                YELLOW(
                    f"Target '{target}' already exists and is not empty\nUse --force to overwrite or choose a different path."
                )
            )
            return 0
        print(CYAN(f"Copying Datatailr demo template to: {target}"))
        username = User.signed_user().name
        copy_tree(src_path, target, username, overwrite=args.force)
        git_init_repo(target)
        print(GREEN("Done. Explore the README.md inside the demo project."))
        install_code_extensions()
        open_folder(target)
        return 0

    elif args.command == "version":
        # Lazy import to avoid unnecessary startup cost
        import importlib.metadata as im

        pkg_version = (
            im.version("datatailr")
            if "datatailr" in {d.metadata["Name"] for d in im.distributions()}
            else "unknown"
        )
        # print nicely colored version info
        print(GREEN(f"Datatailr {pkg_version}"))
        print(CYAN(f"Python {sys.version.split()[0]}"))
        return 0

    # No subcommand: show help and exit
    parser.print_help()
    return 0


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



"""
Copyright (c) 2025 - Datatailr Inc.
All Rights Reserved.

This file is part of Datatailr and subject to the terms and conditions
defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
of this file, in parts or full, via any medium is strictly prohibited.
"""

import os
import sys
import importlib
import subprocess
import inspect
from urllib.parse import urlparse

import numpy as np

try:
    from dt.excel_base import Addin as AddinBase, Queue  # type: ignore
except ImportError as e:
    from datatailr.excel.stubs import AddinBase, Queue


def __progress__(queue, stop):
    from time import sleep

    bar = ["█", "██", "███", "████", "█████", "██████", "███████"]

    count = 0
    while True:
        if stop.is_set():
            return
        queue.push(bar[count % len(bar)])
        count += 1
        sleep(0.25)


def get_package_root(mod):
    # Given module, e.g., dt.excel located at /opt/datatailr/python/dt/excel.py
    # return entry for sys.path so it could be imported as a module.
    # For the module above: /opt/datatailr/python
    mod_path = os.path.abspath(mod.__file__)
    mod_parts = mod.__name__.split(".")
    for _ in range(len(mod_parts)):
        mod_path = os.path.dirname(mod_path)
    return mod_path


def matches_annotation(value, annotation):
    if isinstance(value, np.ndarray):
        return True
    if annotation is bool:
        return isinstance(value, bool) or (type(value) is int and value in (0, 1))
    if annotation is float:
        return isinstance(value, float) or (type(value) is int)
    return isinstance(value, annotation)


def extract_hostname(url: str) -> str | None:
    url = url if url else ""
    if "://" not in url:
        url = "//" + url
    return urlparse(url).hostname


class Addin(AddinBase):
    def __init__(self, *args, **kwargs):
        super(Addin, self).__init__(*args, **kwargs)
        f = inspect.currentframe().f_back
        mod = inspect.getmodule(f)
        if mod is not None:
            setattr(mod, "__dt_addin__", self)

    def run(self, port, ws_port, ide=True):
        # Excel addin executable will try to import an object literally called "addin"
        # from a module passed to dt-excel.sh as an argument. So to find which module
        # to pass to dt-excel.sh, we walk the callstack until a module with "addin"
        # object of type Addin is found. If not -- inform user about this requirement.
        found_module = None
        for frame_info in inspect.stack():
            mod = inspect.getmodule(frame_info.frame)
            if not mod or not hasattr(mod, "__name__"):
                continue

            temp_path = get_package_root(mod)
            sys.path.insert(0, temp_path)
            try:
                imported_mod = importlib.import_module(mod.__name__)
            finally:
                sys.path.pop(0)

            addin_obj = getattr(imported_mod, "__dt_addin__", None)
            if addin_obj is self or id(addin_obj) == id(self):
                found_module = mod
                break

        if not found_module:
            raise ValueError(
                "'__dt_addin__' not found."
            )

        if found_module.__name__ != "__main__":
            # addin.run was called from the initial python script (where __name__ == "__main__")
            module_name = found_module.__name__
            if found_module.__file__ is None:
                raise ValueError(f"Module {found_module.__name__} has no __file__")
            dir_name = os.path.dirname(os.path.abspath(found_module.__file__))
        else:
            # initial python script did not call addin.run() itself (e.g. it imported function that called addin.run)
            filename = inspect.getsourcefile(found_module)
            if filename is None:
                raise ValueError(f"Cannot determine filename for module {found_module}")
            module_name = os.path.splitext(os.path.basename(filename))[0]
            dir_name = os.path.dirname(os.path.abspath(filename))

        ide_flag = "-i" if ide else ""
        hostname = extract_hostname(os.environ.get("VSCODE_PROXY_URI"))

        subprocess.run(
            [
                "bash",
                "-c",
                f'PYTHONPATH="{dir_name}:$PYTHONPATH" /opt/datatailr/bin/dt-excel.sh {ide_flag} -n -H {hostname} -p {port} -w {ws_port} {module_name}',
            ]
        )

    def expose(
        self, description, help, volatile=False, streaming=False, progressbar=False
    ):
        if streaming and progressbar:
            raise ValueError(
                "you cannot specify progressbar and streaming at the same time"
            )

        def decorator(func):
            signature = inspect.signature(func)

            def wrapper(*args, **kwargs):
                # TODO: check whether it's possible to use a kwarg instead so that a decorated function can
                # be called directly from python code without requiring positional argument for _id
                _id = args[0]

                bound = signature.bind_partial(**kwargs)
                bound.apply_defaults()
                for arg in signature.parameters.values():
                    if streaming and arg.name == "queue":
                        continue

                    if not matches_annotation(
                        bound.arguments[arg.name], arg.annotation
                    ):
                        raise ValueError(
                            "excel/python/dt/excel.py: Got argument of wrong type, expected %s or numpy.ndarray, got %s"
                            % (arg.annotation, type(bound.arguments[arg.name]))
                        )
                queue = Queue(self.name.lower() + "." + func.__name__, _id)
                if not streaming:
                    if not progressbar:
                        result = func(**kwargs)
                        if hasattr(result, "tolist"):
                            result = result.tolist()
                        return result

                    from threading import Event, Thread

                    error = None

                    stop = Event()
                    thread = Thread(target=__progress__, args=(queue, stop))
                    thread.start()
                    try:
                        result = func(**kwargs)
                    except Exception as exception:
                        error = str(exception)

                    stop.set()
                    thread.join()

                    if error is not None:
                        queue.error(error)
                    else:
                        queue.push(result)
                    return
                try:
                    func(queue, **kwargs)
                except Exception as exception:
                    queue.error(str(exception))

            self.decorator_impl(
                signature,
                wrapper,
                func.__name__,
                description,
                help,
                volatile,
                streaming or progressbar,
            )
            return wrapper

        return decorator



# *************************************************************************
#
#  Copyright (c) 2025 - Datatailr Inc.
#  All Rights Reserved.
#
#  This file is part of Datatailr and subject to the terms and conditions
#  defined in 'LICENSE.txt'. Unauthorized copying and/or distribution
#  of this file, in parts or full, via any medium is strictly prohibited.
# *************************************************************************
from datatailr import set_allow_unsafe_scheduling

set_allow_unsafe_scheduling(True)


def simple_workflow():
    from datatailr import workflow
    from data_pipelines.data_processing import func_no_args

    @workflow(name="simple_data_pipeline_<>USERNAME<>")
    def simple_data_pipeline():
        func_no_args()

    return simple_data_pipeline()


def simple_app():
    from datatailr import App
    from dashboards.app import main

    app = App(
        name="simple_dashboard_app_<>USERNAME<>",
        entrypoint=main,
        python_requirements="streamlit",
    )
    return app


def simple_service():
    from datatailr import Service
    from services.flask_service import main

    service = Service(
        name="simple_dashboard_service_<>USERNAME<>",
        entrypoint=main,
        python_requirements="flask",
    )
    return service
